package net.sergeych.kiloparsec

import com.ionspin.kotlin.crypto.keyexchange.KeyExchange
import kotlinx.coroutines.CompletableDeferred
import net.sergeych.crypto2.SigningKey
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.Loggable
import net.sergeych.mp_logger.debug
import net.sergeych.utools.unpack

private var serverIds: Int = 0

/**
 * A single kiloparsec server connection. Create it and call [run] and wait until it ends, or cancel
 * parent coroutine context to stop.
 *
 * @param clientInterface local commands, serializable exceptions declarations, etc. usually it is an
 *          reusable object between connection
 * @param device connected device to operate over
 * @param session local session object. Use Unit for no session
 */
class KiloServerConnection<S>(
    private val clientInterface: KiloInterface<S>,
    private val device: Transport.Device,
    private val session: S,
    private val serverSigningKey: SigningKey.Secret? = null
) : RemoteInterface, Loggable by LogTag("SRV${++serverIds}") {

    /**
     * Shortcut to construct with [KiloConnectionData] intance
     */
    @Suppress("unused")
    constructor(localInterface: KiloInterface<S>, connection: KiloConnectionData<S>, serverSecretKey: SigningKey.Secret? = null)
            : this(localInterface, connection.device, connection.session, serverSecretKey)

    private val kiloRemoteInterface = CompletableDeferred<KiloRemoteInterface<S>>()

    /**
     * Run the transport processing loop. This method suspends and only returns when the connection
     * is closed, normally or exceptionally. Cancel the scope where it was called to safely
     * stop processing, otherwise close the [device].
     */
    suspend fun run() {
        val deferredParams = CompletableDeferred<KiloParams<S>>()
        val deferredTransport = CompletableDeferred<Transport<*>>()

        val l0Interface = KiloL0Interface(clientInterface, deferredParams).apply {
            var params: KiloParams<S>? = null
            on(L0Request) {
                val sk = KeyExchange.serverSessionKeys(
                    pair.publicKey, pair.secretKey, it.publicKey
                )

                params = KiloParams(
                    true,
                    deferredTransport.await(),
                    sk,
                    session,
                    null,
                    this@KiloServerConnection
                )

                Handshake(1u, pair.publicKey, serverSigningKey?.seal(params!!.token))
            }

            on(L0ClientId) {
                var p = params ?: throw RemoteInterface.ClosedException("wrong handshake sequence")
                val ci = unpack<ClientIdentity>(p.decrypt(it))
                if( ci.clientIdKey != null ) {
                    if( ci.signature == null )
                        throw RemoteInterface.SecurityException("missing signature")
                    if( !ci.clientIdKey.verify(ci.signature, params!!.token))
                        throw RemoteInterface.SecurityException("wrong signature")
                    p = p.copy(remoteIdentity = ci.clientIdKey)
                    params = p
                }
                deferredParams.complete(p)
                kiloRemoteInterface.complete(
                    KiloRemoteInterface(deferredParams, clientInterface)
                )
                clientInterface.onConnectHandler?.invoke(p.scope)
            }
        }

        val transport = Transport(device, l0Interface, Unit)
            deferredTransport.complete(transport)
        kiloRemoteInterface.complete(KiloRemoteInterface(deferredParams,clientInterface))
        debug { "starting the transport"}
        transport.run()
        debug { "server transport finished" }
    }

    companion object {
        val pair = KeyExchange.keypair()
    }

    override suspend fun <A, R> call(cmd: Command<A, R>, args: A): R {
        return kiloRemoteInterface.await().call(cmd, args)
    }
}