package net.sergeych.kiloparsec.adapter

import io.ktor.client.*
import io.ktor.client.plugins.websocket.*
import io.ktor.http.*
import io.ktor.websocket.*
import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.cancel
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.launch
import net.sergeych.crypto2.SigningKey
import net.sergeych.kiloparsec.KiloClient
import net.sergeych.kiloparsec.KiloConnectionData
import net.sergeych.kiloparsec.KiloInterface
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.exception
import net.sergeych.mp_logger.info
import net.sergeych.mp_logger.warning
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.AtomicCounter

private val counter = AtomicCounter()

fun <S>websocketClient(
    path: String,
    clientInterface: KiloInterface<S> = KiloInterface(),
    client: HttpClient = HttpClient { install(WebSockets) },
    secretKey: SigningKey.Secret? = null,
    sessionMaker: () -> S,
): KiloClient<S> {
    var u = Url(path)
    if (u.encodedPath.length <= 1)
        u = URLBuilder(u).apply {
            encodedPath = "/kp"
        }.build()

    return KiloClient(clientInterface, secretKey) {
        val input = Channel<UByteArray>()
        val output = Channel<UByteArray>()
        globalLaunch {
            val log = LogTag("KC:${counter.incrementAndGet()}:$u")
            client.webSocket({
                url.protocol = u.protocol
                url.host = u.host
                url.encodedPath = u.encodedPath
                url.parameters.appendAll(u.parameters)
            }) {
                try {
                    log.info { "connected to server" }
                    launch {
                        for (block in output) {
                            send(block.toByteArray())
                        }
                        log.info { "input is closed, closing the websocket" }
                        cancel()
                    }
                    for (f in incoming) {
                        if (f is Frame.Binary) {
                            input.send(f.readBytes().toUByteArray())
                        } else {
                            log.warning { "ignoring unexpected frame of type ${f.frameType}" }
                        }
                    }
                }
                catch(_:CancellationException) {
                }
                catch(t: Throwable) {
                    log.exception { "unexpected error" to t }
                }
                log.info { "closing connection" }
            }
        }
        val device = ProxyDevice(input,output) { input.close() }
        KiloConnectionData(device, sessionMaker())
    }
}
