package net.sergeych.kiloparsec.adapter

import kotlinx.coroutines.*
import kotlinx.coroutines.channels.BufferOverflow
import kotlinx.coroutines.channels.Channel
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.exception
import net.sergeych.mp_logger.info
import net.sergeych.mp_logger.warning
import java.net.*
import java.util.concurrent.atomic.AtomicInteger

private val counter = AtomicInteger(0)

class JvmNetworkAddress(val inetAddress: InetAddress, override val port: Int) : NetworkAddress {
    override val host: String by lazy { inetAddress.canonicalHostName }
    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other !is JvmNetworkAddress) return false

        if (inetAddress != other.inetAddress) return false
        if (port != other.port) return false

        return true
    }

    val socketAddress: SocketAddress by lazy { InetSocketAddress(inetAddress,port) }

    override fun hashCode(): Int {
        var result = inetAddress.hashCode()
        result = 31 * result + port
        return result
    }

    override fun toString(): String = "$host:$port"
}

class UdpDatagram(override val message: UByteArray, val inetAddress: InetAddress, val port: Int) : Datagram {

    override val address: NetworkAddress by lazy {
        JvmNetworkAddress(inetAddress, port)
    }

}


@OptIn(DelicateCoroutinesApi::class)
class UdpServer(val port: Int) :
    DatagramConnector, LogTag("UDPS:${counter.incrementAndGet()}") {
    private var isClosed = false


    private val deferredSocket = CompletableDeferred<DatagramSocket>()
    private var job: Job? = null

    private suspend fun start() = try {
        coroutineScope {
            val socket = DatagramSocket(port)
            val buffer = ByteArray(16384)
            val packet = DatagramPacket(buffer, buffer.size)
            deferredSocket.complete(socket)
            while (isActive && !isClosed) {
                try {
                    socket.receive(packet)
                    val data = packet.data.sliceArray(0..<packet.length)
                    val datagram = UdpDatagram(data.toUByteArray(), packet.address, packet.port)
                    if (!channel.trySend(datagram).isSuccess) {
                        warning { "packet lost!" }
                        // and we cause overflow that overwrites the oldest
                        channel.send(datagram)
                    }
                } catch (e: Exception) {
                    if (!isClosed)
                        e.printStackTrace()
                    throw e
                }
            }
            info { "closing socket and reception loop" }

        }
    } catch (_: CancellationException) {
        info { "server is closed" }
    } catch (t: Throwable) {
        exception { "unexpected end of server" to t }
    }

    init {
        job = GlobalScope.launch(Dispatchers.IO) {
            start()
        }
    }

    override fun close() {
        if (!isClosed) {
            if (deferredSocket.isCompleted) {
                runCatching {
                    deferredSocket.getCompleted().close()
                }
            }
            isClosed = true
            job?.cancel(); job = null
        }
    }

    private val channel = Channel<Datagram>(2048, BufferOverflow.DROP_OLDEST)
    override val incoming = channel

    override suspend fun send(message: UByteArray, networkAddress: NetworkAddress) {
        networkAddress as JvmNetworkAddress
        withContext(Dispatchers.IO) {
            val packet = DatagramPacket(
                message.toByteArray(), message.size,
                networkAddress.inetAddress, networkAddress.port
            )
            deferredSocket.await().send(packet)
        }
    }
}