package net.sergeych.kiloparsec.adapter

import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.flow.MutableStateFlow
import net.sergeych.crypto2.encodeVarUnsigned
import net.sergeych.crypto2.readVarUnsigned
import net.sergeych.kiloparsec.Transport
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.warning
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.waitFor
import java.net.InetSocketAddress
import java.net.StandardSocketOptions.TCP_NODELAY
import java.nio.ByteBuffer
import java.nio.channels.AsynchronousSocketChannel
import kotlin.coroutines.cancellation.CancellationException
import kotlin.coroutines.suspendCoroutine

private val log = LogTag("ASTD")

/**
 * Prepend block with its size, varint-encoded
 */
private fun encode(block: UByteArray): ByteArray = (encodeVarUnsigned(block.size.toUInt()) + block).toByteArray()

/**
 * Convert asynchronous socket to a [Transport.Device] using non-blocking nio,
 * in a coroutine-effective manner. Note that it runs coroutines to read/write
 * to the socket in a global scope.These are closed when transport is closed
 *  or the socket is closed, for example, by network failure.
 */
suspend fun asyncSocketToDevice(socket: AsynchronousSocketChannel): InetTransportDevice {
    val deferredDevice = CompletableDeferred<InetTransportDevice>()
    globalLaunch {
        coroutineScope {
            val sendQueueEmpty = MutableStateFlow(true)
            val receiving = MutableStateFlow(false)
            fun stop() {
                cancel()
            }
            // We're in block mode, every block we send worth immediate sending, we do not
            // send partial blocks, so:
            socket.setOption(TCP_NODELAY, true)

            // socket input is to be parsed for blocks, so we receive bytes
            // and decode them to blocks
            val input = Channel<UByte>(1024)
            // copy incoming data from the socket to input channel:
            launch {
                val data = ByteArray(1024)
                val inb = ByteBuffer.wrap(data)
                while (isActive) {
                    inb.position(0)
                    val size: Int = suspendCoroutine { continuation ->
                        socket.read(inb, continuation, IntCompletionHandler)
                    }
                    if (size < 0) stop()
                    else {
//                        println("recvd:\n${data.sliceArray(0..<size).toDump()}\n------------------")
                        for (i in 0..<size) input.send(data[i].toUByte())
                    }
                }
            }

            // output is blocks, so we sent transformed, framed blocks:
            val outputBlocks = Channel<UByteArray>()
            // copy from output to socket:
            launch {
                try {
                    while (isActive) {
                        // wait for the first block to send
                        sendQueueEmpty.value = outputBlocks.isEmpty
                        var data = encode(outputBlocks.receive())

                        // now we're sending, so queue state is sending:
                        sendQueueEmpty.value = false

                        // if there are more, take them all (NO_DELAY optimization)
                        while (!outputBlocks.isEmpty)
                            data += encode(outputBlocks.receive())

                        // now send it all together:
                        val outBuff = ByteBuffer.wrap(data)
                        val cnt = suspendCoroutine { continuation ->
                            socket.write(outBuff, continuation, IntCompletionHandler)
                        }
                        // be sure it was all sent
                        if (outBuff.position() != data.size || cnt != data.size) {
                            throw RuntimeException("unexpected partial write")
                        }
                    }
                    // in the case of just breaking out of the loop:
                    sendQueueEmpty.value = true
                } catch (_: ClosedReceiveChannelException) {
                    stop()
                }
            }
            // transport device copes with blocks:
            val inputBlocks = Channel<UByteArray>()
            // decode blocks from a byte channel read from the socket:
            launch {
                try {
                    while (isActive) {
                        receiving.value = !input.isEmpty
                        val size = readVarUnsigned(input)
                        receiving.value = true
                        if (size == 0u) log.warning { "zero size block is ignored!" }
                        else {
                            val block = UByteArray(size.toInt())
                            for (i in 0..<size.toInt()) {
                                block[i] = input.receive()
                            }
                            inputBlocks.send(block)
                        }
                    }
                } catch (_: CancellationException) {
                } catch (_: ClosedReceiveChannelException) {
                    stop()
                }
                receiving.value = false
            }

            // wait until send queue is empty
            suspend fun flush() {
                yield()
                // do not slow down with collect if it is ok by now:
                if (!sendQueueEmpty.value || !outputBlocks.isEmpty)
                    // wait until all output is sent
                    sendQueueEmpty.waitFor { it && outputBlocks.isEmpty }
            }

            val addr = socket.remoteAddress as InetSocketAddress
            deferredDevice.complete(
                InetTransportDevice(inputBlocks, outputBlocks, JvmNetworkAddress(addr.address, addr.port),
                    { flush() }
                ) {
                    yield()
                    // wait until all received data are parsed, but not too long
                    withTimeoutOrNull( 1000 ) {
                        receiving.waitFor { !it }
                    }
                    // graceful close: flush output
                    flush()
                    // then stop it
                    stop()
                }
            )
        }
        globalLaunch { socket.close() }
    }
    return deferredDevice.await()
}
